# Copyright 2021 Frank David Martinez Muñoz
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in 
# the Software without restriction, including without limitation the rights to use, 
# copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is furnished to do so, 
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in 
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
# INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A 
# PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION 
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from bawr import utils
from pathlib import Path
import re
import os

RE_VAR = re.compile(r'\$\{(.*?)\}')

class CppEmbedded:

    # Config
    name = None
    namespace = None
    source = None

    # Generated
    instance = None
    output_data = None
    output_header = None

    def build(self, env):
        print("[Cpp Bin Embedder] Start ...")

        def replacer(match):
            return str(getattr(env, match.group(1), match.group(1)))

        input_bin = Path(RE_VAR.sub(replacer, self.source).replace('/', os.sep).replace('\\', os.sep))
        name = self.name if self.name else input_bin.stem
        self.name = name

        self.output_data = Path(env.BAWR_OUTPUT_DIR, name + ".cpp")
        self.output_header = Path(env.BAWR_OUTPUT_DIR, name + ".hpp")

        utils.bin_to_cpp_data(
            self.output_data, 
            self.namespace or 'icons', 
            input_bin
        )




# Copyright 2021 Frank David Martinez Muñoz
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in 
# the Software without restriction, including without limitation the rights to use, 
# copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is furnished to do so, 
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in 
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
# INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A 
# PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION 
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import re

RE_GLYPH_NAME = re.compile(r'^[a-zA-Z](\w)*$')
RE_GLYPH_BADCHAR = re.compile(r'\W+')

def get_glyph_name(name):
    glyph_name = RE_GLYPH_BADCHAR.sub("_", name)
    if not RE_GLYPH_NAME.match(glyph_name):
        print("[error] Invalid glyph name {} from {}".format(glyph_name, name))
    return glyph_name

class Glyph:
    def __init__(self, code, name, path, source):
        self.code = code
        self.name = get_glyph_name(name)
        self.path = path
        self.source = source

    def __repr__(self):
        return "{}:{}:{:x} ({})".format(self.source.__name__, self.name, self.code, self.path)




# Copyright 2021 Frank David Martinez Muñoz
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in 
# the Software without restriction, including without limitation the rights to use, 
# copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is furnished to do so, 
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in 
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
# INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A 
# PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION 
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import shutil
from pathlib import Path
import sys


def get_command_path(path, name):
    if not path:
        return Path(shutil.which(name))
    elif isinstance(path, str):
        return Path(path)


class Environment:
    FONTFORGE_PATH = None
    INKSCAPE_PATH = None
    BAWR_OUTPUT_DIR = None
    BAWR_SOURCE_DIR = None
    BAWR_CONFIG_FILE = None

    def build(self, cfg_file, src_dir, out_dir):
        self.FONTFORGE_PATH = get_command_path(self.FONTFORGE_PATH, 'fontforge')
        self.INKSCAPE_PATH = get_command_path(self.INKSCAPE_PATH, 'inkscape')

        if src_dir:
            self.BAWR_SOURCE_DIR = Path(src_dir)
        elif not self.BAWR_SOURCE_DIR:
            self.BAWR_SOURCE_DIR = Path('.')
        else:
            self.BAWR_SOURCE_DIR = Path(self.BAWR_SOURCE_DIR)

        if out_dir:
            self.BAWR_OUTPUT_DIR = Path(out_dir)
        elif not self.BAWR_OUTPUT_DIR:
            self.BAWR_OUTPUT_DIR = Path('.', 'build')
        else:
            self.BAWR_OUTPUT_DIR = Path(self.BAWR_OUTPUT_DIR)
            
        self.BAWR_CONFIG_FILE = Path(cfg_file)
        if not self.BAWR_CONFIG_FILE.exists():
            print(f"Configuration file not found: {cfg_file}")
            sys.exit(-1)

        for k, v in self.__dict__.items():
            print("ENV: {} = {}".format(k, v))



# Copyright 2021 Frank David Martinez Muñoz
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in 
# the Software without restriction, including without limitation the rights to use, 
# copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is furnished to do so, 
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in 
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
# INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A 
# PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION 
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import re
import struct
import os
from pathlib import Path
from collections.abc import Iterable

class StringReplacer:
    '''
    table = {
        "aaa"    : "[This is three a]",
        "b+"     : "[This is one or more b]",
        r"<\w+>" : "[This is a tag]"
    }
    replacer = StringReplacer(table, True)
    sample1 = "whatever bb, aaa, <star> BBB <end>"
    print(replacer(sample1))
    >> whatever [This is one or more b], [This is three a], [This is a tag] [This is one or more b] [This is a tag]
    '''
    def __init__(self, replacements, ignore_case=False):
        patterns = sorted(replacements, key=len, reverse=True)
        self.replacements = [replacements[k] for k in patterns]
        re_mode = re.IGNORECASE if ignore_case else 0
        self.pattern = re.compile('|'.join(("({})".format(p) for p in patterns)), re_mode)
        self.changed = False
        def tr(matcher):
            self.changed = True
            index = next((index for index,value in enumerate(matcher.groups()) if value), None)
            return self.replacements[index]
        self.tr = tr

    def __call__(self, string):
        self.changed = False
        return self.pattern.sub(self.tr, string)


def bin_to_cpp_data(cpp_file, namespace, bin_file, data_var='DATA', size_var='SIZE', sub_namespace='data'):

    if not bin_file.exists():
        print(f"[Cpp Bin Embedder] Bin file not found: {bin_file}\n")
        return   

    hpp_file = Path(cpp_file.parent, cpp_file.stem + ".hpp")
    print(f"[Cpp Bin Embedder] < {bin_file}")
    print(f"[Cpp Bin Embedder] > {cpp_file}")
    print(f"[Cpp Bin Embedder] > {hpp_file}")

    padding = [None, b"\x00", b"\x00\x00", b"\x00\x00\x00"]
    with open(hpp_file, 'w') as f:
        f.write("#pragma once\n\n")
        f.write(f"namespace {namespace} {{\n")
        f.write(f" namespace {sub_namespace} {{\n")
        f.write(f"  extern const unsigned int {size_var};\n")
        f.write(f"  extern const unsigned int {data_var}[];\n")
        f.write(" }\n")
        f.write("}\n")

    with open(cpp_file, 'w') as f:
        f.write(f'#include "{hpp_file.name}"\n')
        f.write(f"namespace {namespace} {{\n")
        f.write(f" namespace {sub_namespace} {{\n")
        f.write(f"  const unsigned int {size_var} = {os.stat(bin_file).st_size};\n")
        f.write(f"  const unsigned int {data_var}[] = {{\n  ")
        with open(bin_file, 'rb') as bin:
            n = 0
            c = bin.read(4)
            while (c):
                val = struct.unpack("<L", c)[0]
                f.write(f"{val:#010x}")
                f.write(",")
                n += 1
                if n % 12 == 0:
                    f.write("\n  ")
                c = bin.read(4)
                missing = 4 - len(c)
                if missing > 0 and missing < 4:
                    c += padding[missing]
        f.write("};\n")
        f.write(" }\n")
        f.write("}\n")

def get_cache_dir(env):
    path = Path(env.BAWR_OUTPUT_DIR, "__bawr_cache__")
    if not path.exists():
        os.makedirs(path, exist_ok=True)
    return path

def as_iterable(obj):
    if isinstance(obj, Iterable):
        return obj
    return (obj,)



# Copyright 2021 Frank David Martinez Muñoz
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in 
# the Software without restriction, including without limitation the rights to use, 
# copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is furnished to do so, 
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in 
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
# INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A 
# PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION 
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

class IconSet:

    # Config
    src = None
    select = []
    options = None
    
    # Generated
    name = None
    instance = None

    def build(self, env):
        self.__class__.name = self.__class__.__name__
        print("[Icon Set] {}: {}".format(self.name, self.src))


# Copyright 2021 Frank David Martinez Muñoz
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in 
# the Software without restriction, including without limitation the rights to use, 
# copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is furnished to do so, 
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in 
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
# INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A 
# PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION 
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from pathlib import Path
from datetime import datetime

class CppFontHeader:

    # Config
    name = None
    namespace = None
    constexpr = False
    macros = True
    macro_prefix = 'Icon_'
    source = None

    # Generated
    instance = None
    output = None

    def build(self, env):
        print("[CPP Font Header] Start ...")

        if self.source is None or self.source.instance is None:
            print(f"[CPP Font Header] Invalid source (CppFontHeader.source)")
            return

        instance = self.source.instance
        icons = instance.icons
        build_dir = env.BAWR_OUTPUT_DIR
        self.namespace = self.namespace or 'icons'
        namespace = self.namespace

        header_file = Path(build_dir, self.name or instance.name + "_codes.hpp")
        self.output = header_file
        print(f"[CPP Font Header] {str(header_file)}")
        with open(header_file, 'w') as f:
            f.write(f'#pragma once\n\n')
            f.write(f'// Generated: {datetime.now()}\n\n')

            if self.macros:
                f.write(f'#define {self.macro_prefix}{"Font_Family":<32} "{instance.family}"\n')
                f.write(f'#define {self.macro_prefix}{"Font_StartCode":<32} {hex(instance.start_code)}\n')
                f.write(f'#define {self.macro_prefix}{"Font_EndCode":<32} {hex(instance.end_code)}\n')
                for glyph in icons:
                    if glyph.code > 0:
                        code = "U+" + hex(glyph.code)[2:]
                        literal = repr( chr(glyph.code).encode( 'utf-8' ))[ 2:-1 ]
                        f.write(f'#define {self.macro_prefix}{glyph.name:<32} "{literal}" //< {code}\n')

            if self.constexpr:
                constexpr = 'constexpr'
            else:
                constexpr = 'const'

            f.write(f'\nnamespace {namespace}\n{{\n')       
            f.write(f'    {constexpr} auto {"Font_Family":<32} = "{instance.family}";\n')
            f.write(f'    {constexpr} auto {"Font_StartCode":<32} = {hex(instance.start_code)};\n')
            f.write(f'    {constexpr} auto {"Font_EndCode":<32} = {hex(instance.end_code)};\n')
            for glyph in icons:
                if glyph.code > 0:
                    code = "U+" + hex(glyph.code)[2:]
                    literal = repr( chr(glyph.code).encode( 'utf-8' ))[ 2:-1 ]
                    f.write(f'    {constexpr} auto {glyph.name:<32} = "{literal}"; //< {code}\n')

            f.write(f'\n    namespace code\n    {{\n')       
            for glyph in icons:
                if glyph.code > 0:
                    code = "U+" + hex(glyph.code)[2:]
                    literal = repr( chr(glyph.code).encode( 'utf-8' ))[ 2:-1 ]
                    f.write(f'        {constexpr} auto {glyph.name:<32} = "{code}"; //< {literal}\n')
            f.write('    }\n')

            f.write('}\n')




# Copyright 2021 Frank David Martinez M.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in 
# the Software without restriction, including without limitation the rights to use, 
# copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is furnished to do so, 
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in 
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
# INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A 
# PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION 
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import fontforge, psMat
import os, sys, math

# +-------------------------------------------------------------------+
# | Start Generated Glyphs                                            |
# | defines: config, select, options, build_dir, global_transform,    |
# |          output_formats, verbose                                  |

#@generated

# | End Generated Glyphs                                              |
# +-------------------------------------------------------------------+

try:
    if not os.path.exists(build_dir):
        os.mkdir(build_dir)
except:
    sys.exit("Error creating directory: %s" % build_dir)   


def parse_transformation(t):
    '''
        ('t|translate', x, y)
        ('s|scale', x, y)
        ('r|rotate', deg)
        ('x|skew', deg)
    '''
    m = None
    if t:
        m = psMat.identity()
        for tx in t:
            if tx[0] == 't' or tx[0] == 'translate':
                m = psMat.compose(m, psMat.translate(tx[1], tx[2]))
            elif tx[0] == 'r' or tx[0] == 'rotate':
                m = psMat.compose(m, psMat.rotate(math.radians(tx[1])))
            elif tx[0] == 's' or tx[0] == 'scale':
                m = psMat.compose(m, psMat.scale(tx[1], tx[2]))
            elif tx[0] == 'x' or tx[0] == 'skew':
                m = psMat.compose(m, psMat.skew(math.radians(tx[1])))
    return m


font = fontforge.font()
font.encoding = "UnicodeFull"
font.copyright = config.font_copyright
font.familyname = config.font_family
font.fontname = config.font_name
font.fullname = config.font_name

for code, name, path, options in select:
    glyph = font.createChar(code, name)
    glyph.importOutlines(path)
    if options:
        lt = parse_transformation(options.get('font_transformation', None)) 
        if lt:
            glyph.transform(lt)
    if global_transform:
        glyph.transform(global_transform)

    if verbose:
        print("\n%s:\n  file: %s\n  code: %s" % (name, path, hex(code)))

for format in output_formats:
    font.generate('%s/%s.%s' % (build_dir, font.fontname, format))
